#ifdef __cplusplus
    #include <cstdlib>
#else
    #include <stdlib.h>
#endif
#include <string>
#ifdef _MSC_VER
#include <SDL.h>
#else
#include <SDL2/SDL.h>
#endif
#include <libssh/libssh.h>
#ifndef _MSC_VER
#include <wait.h>
#endif
#include <sys/stat.h>
#include <limits.h>
#include "Terminal.h"
#include "magiterm.h"
#include "led.h"
#include "icon.h"
#include "mtermans.h"

#ifdef _MSC_VER
#define PATH_MAX MAX_PATH
#endif

ssh_channel chan = NULL;
SDL_Thread *dlthread;

extern void zmodem_upload(char *filename);

SDL_Surface *loadImage(unsigned char *iaddr, int ilen) {
    SDL_RWops *src = NULL;
    SDL_Surface *s0;

    src = SDL_RWFromMem(iaddr, ilen);
    s0 = SDL_LoadBMP_RW(src, 1);

    return s0;
}

int main ( int argc, char** argv )
{
    FILE *fptr;
    unsigned char c;
    int state = 0;
    char host[256];
    char username[256];
    char password[256];
    char buffer[256];
    char uploadfile[PATH_MAX];
    int i = 0;
    int port = 0;
    int pos = 0;
    SDL_Window *window;
    SDL_Surface *toolbox;
    SDL_Surface *led;
    SDL_Surface *icon;
    int sshstate;
    size_t hlen;
    unsigned char *hash = NULL;
    char *hexa;
    char buf[10];
    ssh_session sess;

    int verbosity = SSH_LOG_TRACE;
    SDL_Rect dest;
    int rz = 0;
    int sz = 0;
    int zmodem = 0;
    ssh_key key;
    struct stat st;
    // initialize SDL video
    if ( SDL_Init( SDL_INIT_EVERYTHING ) < 0 )
    {
        printf( "Unable to init SDL: %s\n", SDL_GetError() );
        return 1;
    }


    // make sure SDL cleans up before exit
    atexit(SDL_Quit);

    // create a new window
    window = SDL_CreateWindow("MagiTerm v0.04", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, 640, 480, SDL_WINDOW_SHOWN);
    SDL_Surface* screen = SDL_GetWindowSurface(window);
    if ( !screen )
    {
        printf("Unable to set 640x480 video: %s\n", SDL_GetError());
        return 1;
    }

    // load an image

    toolbox = loadImage((unsigned char *)TOOLBOX_IMAGE_DATA, sizeof(TOOLBOX_IMAGE_DATA));
    led = loadImage((unsigned char *)LED_IMAGE_DATA, sizeof(LED_IMAGE_DATA));
    icon = loadImage((unsigned char *)ICON_IMAGE_DATA, sizeof(ICON_IMAGE_DATA));

    SDL_SetWindowIcon(window, icon);


    Terminal *term = new Terminal();

    term->twrite((unsigned char *)MAGITERM_ANSI_DATA, sizeof(MAGITERM_ANSI_DATA));
    term->twrite((unsigned char *)"\r\n\x1b[0m", 6);
    
    if (argc > 1) {
        strncpy(host, argv[1], 256);
        state = 2;
        if (argc > 2) {
            port = atoi(argv[2]);
        } else {
            port = 22;
        }
        term->twrite((unsigned char *)"USER: ", 6);
    } else {
        term->twrite((unsigned char *)"HOST: ", 6);
    }

    // program main loop
    bool done = false;

    SDL_StartTextInput();

    while (!done)
    {
        // message processing loop
        SDL_Event event;
        
        while (SDL_PollEvent(&event))
        {
            // check for messages
            switch (event.type)
            {
                // exit if the window is closed
            case SDL_QUIT:
                done = true;
                break;

            case SDL_KEYDOWN:
                {
                    // exit if ESCAPE is pressed
                    if (event.key.keysym.sym == SDLK_x && SDL_GetModState() & KMOD_ALT) {
                        done = true;
                    }

                    if (SDL_GetModState() & KMOD_CTRL && state == 9) {
                        if (event.key.keysym.sym >= 'a' && event.key.keysym.sym <= 'z') {
                            snprintf(buf, 10, "%c", event.key.keysym.sym  - 'a' + 1);
                            ssh_channel_write(chan, buf, 1);
                            sz = 1;
                        }
                    }

                    if( event.key.keysym.sym == SDLK_BACKSPACE && pos > 0 ) {
                        //lop off character
                        if (state == 0) {
                            host[pos--];
                            host[pos] = '\0';
                        } else if (state == 1) {
                            port = port / 10;
                            pos--;
                        } else if (state == 2) {
                            username[pos--];
                            username[pos] = '\0';
                        } else if (state == 3) {
                            password[pos--];
                            password[pos] = '\0';
                        } else if (state == 6) {
                            buf[pos--];
                            buf[pos] = '\0';
                        } else if (state == 13) {
                            uploadfile[pos--];
                            uploadfile[pos] = '\0';
                        }
                        term->twrite((unsigned char *)"\x1b[D \x1b[D", 7);
                    } else if (event.key.keysym.sym == SDLK_BACKSPACE && state == 9) {
                        ssh_channel_write(chan, "\b", 1);
                        sz = 1;
                    } 

                    if (event.key.keysym.sym == SDLK_RETURN) {
                        if (state == 0 && pos != 0) {
                            state = 1;
                            pos = 0;
                            term->twrite((unsigned char *)"\r\nPORT: ", 8); 
                        } else if (state == 1) {
                            if (pos == 0) port = 22;
                            state = 2;
                            pos = 0;
                            term->twrite((unsigned char *)"\r\nUSER: ", 8); 
                        } else if (state == 2 && pos != 0) {
                            state = 3;
                            pos = 0;
                            term->twrite((unsigned char *)"\r\nPASS: ", 8); 
                        } else if (state == 3 && pos != 0) {
                            state = 4;
                            pos = 0;
                            term->twrite((unsigned char *)"\r\nCONNECTING...\r\n", 17);
                        } else if (state == 6 && pos != 0) {
                            state = 7;
                            pos = 0;
                        } else if (state == 9) {
                            ssh_channel_write(chan, "\r", 1);
                            sz = 1;
                        } else if (state == 13) {
                            // upload file
                            if (stat(uploadfile, &st) != 0) {
                                term->twrite((unsigned char *)"\r\nFILE NOT FOUND...\r\n", 21);
                                state = 9;
                                pos = 0;
                            } else {
                                term->twrite((unsigned char *)"\r\nUPLOADING...\r\n", 16);
                                pos = 0;
                                state = 11;
                                zmodem_upload(uploadfile);
                            }
                        }
                    }

                    if (state == 13 && event.key.keysym.sym == SDLK_ESCAPE) {
                        term->twrite((unsigned char *)"\r\nFILE UPLOAD CANCELED...\r\n", 27);
                        state = 9;
                    }

                    if (state == 9) {
                        if (event.key.keysym.sym == SDLK_ESCAPE) {
                            ssh_channel_write(chan, "\x1b", 1);
                            sz = 1;
                        }
                        if (event.key.keysym.sym == SDLK_UP) {
                            ssh_channel_write(chan, "\x1b[A", 3);
                            sz = 1;
                        }
                        if (event.key.keysym.sym == SDLK_DOWN) {
                            ssh_channel_write(chan, "\x1b[B", 3);
                            sz = 1;
                        }     
                        if (event.key.keysym.sym == SDLK_LEFT) {
                            ssh_channel_write(chan, "\x1b[D", 3);
                            sz = 1;
                        }
                        if (event.key.keysym.sym == SDLK_RIGHT) {
                            ssh_channel_write(chan, "\x1b[C", 3);
                            sz = 1;
                        }                             
                        if (event.key.keysym.sym == SDLK_END) {
                            ssh_channel_write(chan, "\x1b[K", 3);
                            sz = 1;
                        }
                        if (event.key.keysym.sym == SDLK_HOME) {
                            ssh_channel_write(chan, "\x1b[H", 3);
                            sz = 1;
                        }                        
                        if (event.key.keysym.sym == SDLK_PAGEUP) {
                            ssh_channel_write(chan, "\x1b[V", 3);
                            sz = 1;
                        }                        
                        if (event.key.keysym.sym == SDLK_PAGEDOWN) {
                            ssh_channel_write(chan, "\x1b[U", 3);
                            sz = 1;
                        }
                        if (event.key.keysym.sym == SDLK_F1) {
                            ssh_channel_write(chan, "\x1b[OP", 4);
                            sz = 1;
                        }                    
                        if (event.key.keysym.sym == SDLK_F2) {
                            ssh_channel_write(chan, "\x1b[OQ", 4);
                            sz = 1;
                        }                    
                        if (event.key.keysym.sym == SDLK_F3) {
                            ssh_channel_write(chan, "\x1b[OR", 4);
                            sz = 1;
                        }                    
                        if (event.key.keysym.sym == SDLK_F4) {
                            ssh_channel_write(chan, "\x1b[OS", 4);
                            sz = 1;
                        }                                            
                        if (event.key.keysym.sym == SDLK_F5) {
                            ssh_channel_write(chan, "\x1b[[15~", 6);
                            sz = 1;
                        }                    
                        if (event.key.keysym.sym == SDLK_F6) {
                            ssh_channel_write(chan, "\x1b[[17~", 6);
                            sz = 1;
                        }                    
                        if (event.key.keysym.sym == SDLK_F7) {
                            ssh_channel_write(chan, "\x1b[[18~", 6);
                            sz = 1;
                        }                    
                        if (event.key.keysym.sym == SDLK_F8) {
                            ssh_channel_write(chan, "\x1b[[19~", 6);
                            sz = 1;
                        }                    
                        if (event.key.keysym.sym == SDLK_F9) {
                            ssh_channel_write(chan, "\x1b[[20~", 6);
                            sz = 1;
                        }                    
                        if (event.key.keysym.sym == SDLK_F10) {
                            ssh_channel_write(chan, "\x1b[[21~", 6);
                            sz = 1;
                        }                    
                        if (event.key.keysym.sym == SDLK_F11) {
                            ssh_channel_write(chan, "\x1b[[23~", 6);
                            sz = 1;
                        }                    
                        if (event.key.keysym.sym == SDLK_F12) {
                            ssh_channel_write(chan, "\x1b[[24~", 6);
                            sz = 1;
                        }
                    }
                }
                break;
            case SDL_TEXTINPUT:
                {
                    if (state == 0) {
                        if (strlen(host) + strlen(event.text.text) > 255) {
                            state = 1;
                            pos = 0;
                            term->twrite((unsigned char *)"\r\nPORT: ", 8);
                        } else {
                            for (i=0;i<strlen(event.text.text);i++) {
                                host[pos++] = event.text.text[i];
                                host[pos] = '\0';
                                term->twrite((unsigned char *)&event.text.text[i], 1);
                            }
                            
                        }
                    } else if (state == 1) {
                        for (i=0;i<strlen(event.text.text);i++) {
                            if (event.text.text[i] >= '0' && event.text.text[i] <= '9') {
                                pos++;
                                port = port * 10 + (event.text.text[i] - '0');
                                term->twrite((unsigned char *)&event.text.text[i], 1);
                            }
                        }                     
                    } else if (state == 2) {
                        if (strlen(username) + strlen(event.text.text) > 255) {
                            state = 3;
                            pos = 0;
                            term->twrite((unsigned char *)"\r\nPASS: ", 8);
                        } else {
                            for (i=0;i<strlen(event.text.text);i++) {
                                username[pos++] = event.text.text[i];
                                username[pos] = '\0';
                                term->twrite((unsigned char *)&event.text.text[i], 1);
                            }
                            
                        }
                    } else if (state == 3) {
                        if (strlen(password) + strlen(event.text.text) > 255) {
                            state = 4;
                            pos = 0;
                            term->twrite((unsigned char *)"\r\n\r\nCONNECTING...\r\n", 19);
                        } else {
                            for (i=0;i<strlen(event.text.text);i++) {
                                password[pos++] = event.text.text[i];
                                password[pos] = '\0';
                                term->twrite((unsigned char *)"*", 1);
                            }
                        }
                    } else if (state == 6) {
                        if (strlen(buf) + strlen(event.text.text) >  9) {
                            state = 7;
                            pos = 0;
                        } else {
                            for (i=0;i<strlen(event.text.text);i++) {
                                buf[pos++] = event.text.text[i];
                                buf[pos] = '\0';
                                term->twrite((unsigned char *)&event.text.text[i], 1);
                            }     
                        }
                    } else if (state == 9) {
                       if (!(SDL_GetModState() & KMOD_CTRL)) {
                            if (!ssh_channel_is_eof(chan) && ssh_channel_is_open(chan)) {
                                sz = 1;
                                i = ssh_channel_write(chan, event.text.text, strlen(event.text.text));
                                if (i != strlen(event.text.text)) {
                                    state = 10;
                                }
                            }
                        }
                    } else if (state == 13) {

                        for (i=0;i<strlen(event.text.text);i++) {
                            if (pos != PATH_MAX - 1) {                   
                                uploadfile[pos++] = event.text.text[i];
                                uploadfile[pos] = '\0';
                                term->twrite((unsigned char *)&event.text.text[i], 1);
                            }
                        }   
                    }
                }
            } // end switch
        } // end of message processing

        // DRAWING STARTS HERE

        // clear screen
        SDL_FillRect(screen, 0, SDL_MapRGB(screen->format, 0, 0, 0));
        SDL_BlitSurface(term->getSurface(), 0, screen, 0);

        dest.x = 0;
        dest.y = 400;
        dest.w = 640;
        dest.h = 80;

        SDL_BlitSurface(toolbox, 0, screen, &dest);

        if (sz) {
            dest.x = 561;
            dest.y = 413;
            dest.w = 23;
            dest.h = 23;

            SDL_BlitSurface(led, 0, screen, &dest);
        }
        if (rz) {
            dest.x = 561;
            dest.y = 439;
            dest.w = 23;
            dest.h = 23;

            SDL_BlitSurface(led, 0, screen, &dest);
        }

        // DRAWING ENDS HERE
        rz = 0;
        sz = 0;
        // finally, update the screen :)
        SDL_UpdateWindowSurface(window);

        if (state == 4) {
            sess = ssh_new();
            if (sess == NULL) {
                term->twrite((unsigned char *)"Unable to create SSH Session!", 29);
                state = -1;
            }
            ssh_options_set(sess, SSH_OPTIONS_HOST, host);
            ssh_options_set(sess, SSH_OPTIONS_PORT, &port);
            ssh_options_set(sess, SSH_OPTIONS_USER, username);
           // ssh_options_set(sess, SSH_OPTIONS_LOG_VERBOSITY, &verbosity);
            i = ssh_connect(sess);

            if (i != SSH_OK) {
                term->twrite((unsigned char *)"Unable to connect!", 18);
                ssh_free(sess);
                state = -1;
            } else {
                state = 5;
            }
        } else if (state == 5) {
            sshstate = ssh_is_server_known(sess);
            if (ssh_get_server_publickey(sess, &key) != SSH_OK) {
                exit(-1);
            }
            
            ssh_get_publickey_hash(key, SSH_PUBLICKEY_HASH_SHA1, &hash, &hlen);

            ssh_key_free(key);

            if (hlen < 0) {
                ssh_disconnect(sess);
                ssh_free(sess);
                state = -1;
            } else {
                switch (sshstate) {
                    case SSH_SERVER_KNOWN_OK:
                        state = 8;
                        break; /* ok */
                    case SSH_SERVER_KNOWN_CHANGED:
                        hexa = ssh_get_hexa(hash, hlen);
                        term->twrite((unsigned char *)"Host key for server has changed! It is now:\r\n", 45);
                        term->twrite((unsigned char *)hexa, strlen(hexa));
                        term->twrite((unsigned char *)"\r\nFor security reasons, connection will be stopped\r\n", 52);
                        ssh_clean_pubkey_hash(&hash);
                        ssh_string_free_char(hexa);
                        ssh_disconnect(sess);
                        ssh_free(sess);
                        state = -1;
                        break;
                    case SSH_SERVER_FOUND_OTHER:
                        term->twrite((unsigned char *)"The host key for this server was not found but an other type of key exists.\r\n", 77);
                        term->twrite((unsigned char *)"An attacker might change the default server key to confuse your client into thinking the key does not exist\r\n", 109);
                        ssh_clean_pubkey_hash(&hash);
                        ssh_disconnect(sess);
                        ssh_free(sess);
                        state = -1;
                        break;
                    case SSH_SERVER_FILE_NOT_FOUND:
                        term->twrite((unsigned char *)"Could not find known host file.\r\n", 33);
                        term->twrite((unsigned char *)"If you accept the host key here, the file will be automatically created.\r\n", 74);
                        /* fallback to SSH_SERVER_NOT_KNOWN behavior */
                    case SSH_SERVER_NOT_KNOWN:
                        hexa = ssh_get_hexa(hash, hlen);
                        term->twrite((unsigned char *)"The server is unknown. Do you trust the host key?\r\n", 51);
                        term->twrite((unsigned char *)"Public key hash: ", 17);
                        term->twrite((unsigned char *)hexa, strlen(hexa));
                        term->twrite((unsigned char *)"\r\n: ", 4);
                        ssh_string_free_char(hexa);
                        ssh_clean_pubkey_hash(&hash);
                        state = 6;
                        break;
                    case SSH_SERVER_ERROR:
                        term->twrite((unsigned char *)"Server Error!\r\n", 15);
                        ssh_clean_pubkey_hash(&hash);
                        ssh_disconnect(sess);
                        ssh_free(sess);
                        state = -1;
                        break;
                }
            }
        } else if (state == 7) {
#ifndef _MSC_VER
            if (strcasecmp(buf, "yes") == 0) {
#else
			if (stricmp(buf, "yes") == 0) {
#endif
                if (ssh_write_knownhost(sess) < 0) {
                    term->twrite((unsigned char *)"Error!\r\n", 8);
                    ssh_disconnect(sess);
                    ssh_free(sess);
                    state = -1;
                } else {
                    state = 8;
                }
            } else {
                term->twrite((unsigned char *)"Disconnected!\r\n", 15);
                ssh_disconnect(sess);
                ssh_free(sess);
                state = -1;                
            }
        } else if (state == 8) {
            i = ssh_userauth_password(sess, username, password);
            if (i != SSH_AUTH_SUCCESS) {
                term->twrite((unsigned char*)"Access Denied!\r\n", 16);
                ssh_disconnect(sess);
                ssh_free(sess);
                state = -1; 
            } else {
                chan = ssh_channel_new(sess);
                if (chan == NULL) {
                    term->twrite((unsigned char*)"Couldn't create channel!\r\n", 26);
                    ssh_disconnect(sess);
                    ssh_free(sess);
                    state = -1;     
                } else {
                    i = ssh_channel_open_session(chan);
                    if (i != SSH_OK) {
                        term->twrite((unsigned char *)"Couldn't open session!\r\n", 23);
                        ssh_channel_send_eof(chan);
                        ssh_channel_close(chan);
                        ssh_channel_free(chan);                        
                        ssh_disconnect(sess);
                        ssh_free(sess);
                        state = -1;                              
                    } else {
                        i = ssh_channel_request_pty_size(chan, "syncterm", 80, 25);
                        if (i != SSH_OK) {
                            term->twrite((unsigned char*)"Couldn't set request PTY!\r\n", 27);
                            ssh_channel_send_eof(chan);
                            ssh_channel_close(chan);
                            ssh_channel_free(chan);                        
                            ssh_disconnect(sess);
                            ssh_free(sess);
                            state = -1;
                        } else {             
                            i = ssh_channel_request_shell(chan);
                            if (i != SSH_OK) {
                                term->twrite((unsigned char*)"Couldn't request shell!\r\n", 25);
                                ssh_channel_send_eof(chan);
                                ssh_channel_close(chan);
                                ssh_channel_free(chan);                        
                                ssh_disconnect(sess);
                                ssh_free(sess);
                                state = -1;
                            } else {
                                term->setZmodem(&zmodem);
                                state = 9;
                            }
                        }
                    }
                }
            }
        } else if (state == 9) {
            i = ssh_channel_read_nonblocking(chan, buffer, 256, 0);
            if (i > 0) {
                rz = 1;
                term->twrite((unsigned char *)buffer, i);
                if (zmodem == 1) {
                    state = 11;
                } else if (zmodem == 2) {
                    state = 12;
                }
            } else if (i < 0) {
                state = 10;
            }
        } else if (state == 10) {
            term->twrite((unsigned char *)"\r\n\r\n\r\nDisconnected...\r\n", 23);
            ssh_channel_send_eof(chan);
            ssh_channel_close(chan);            
            ssh_channel_free(chan);                        
            ssh_disconnect(sess);
            ssh_free(sess);
            state = -1;
        } else if (state == 11) {
            if (!zmodem) {
                SDL_WaitThread(dlthread, NULL);
                state = 9;
            }
        } else if (state == 12) {
            // get upload filename....
            term->twrite((unsigned char *)"\r\n\r\nFILE TO UPLOAD: ", 20);

            state = 13;
        }
#ifndef _MSC_VER
        usleep(3);
#endif
    } // end main loop

    // free loaded bitmap
    SDL_FreeSurface(toolbox);
    SDL_FreeSurface(led);
    SDL_FreeSurface(icon);
    // all is well ;)
    printf("Exited cleanly\n");
    return 0;
}
